#include <R.h>
#include <Rinternals.h>
#include <nag.h>
#include <nage04.h>
#include "nag_rwrapper_utils.h"

#ifdef __cplusplus
extern "C" {
#endif
  void NAG_CALL we04abcF(double xc, double *fc, Nag_Comm *comm);
  void NAG_CALL we04abcL(double xc, double *fc, Nag_Comm *comm);
#ifdef __cplusplus
}
#endif


SEXP we04abc(SEXP funct, SEXP ab,  SEXP e, SEXP mfun, SEXP rho) {
  /* we04abc: Wrapper for the NAG C library routine e04abc.
     On entry:
     funct  - Either an R function or the body of a function which we wish to minimise
              The R function pointed to by funct can either be a function, or
              the body of a function, i.e. if
        
              ff = function(y) {sin(y) / y}

              then can either pass ff directly, or could pass
                     {as.double(ff(x))}
              or
               gg = function(x) {
                  as.double(ff(x))
               }
              and then use body(gg)

              In the case of the code segment, or using the body function, the variable x 
              must be used to denote the value at which the function is to be evaluated.

              Normally would only use one of the above, and hence could remove the unused
              code from this function

     ab     - Vector of two elements containing the interval holding the minimum, 
              this is equivalent to c(a,b), where "c" is the R function of same name
              and "a" and "b" are arguments to e04abc.
     e      - Vector of two elements containing the relative and absolute errors. This is
              equivalent to c(e1,e2), where "e1" and "e2" are arguments to e04abc. If
              e = c(NA,NA) the default values are used for e1 and e2, similarly if e only
              has one value then a default is used for e2.
     mfun   - Integer holding the maximum number of function evaluations (equivalent to
              c(max_fun), where "max_fun" is an argument to e04abc. If mfun = c(NA), then
              a default value of 100 is used.
     rho    - R environment used to evaluate the function

     Returns:
     An R list, with elements "x", "f" and "ab", where "x" and "f" are equivalent to the
     output arguements of e04abc of the same name (i.e. the estimated minimum and the 
     value of the function at x), and "ab" is the equivalent to c(a,b) on exit from e04abc, i.e.
     the interval containing the estimated minimum.
  */
  double e1, e2, a, b, t, x, f;
  Integer max_fun, ifail;
  NagError fail;
  Nag_Comm comm;
  void **p;
  SEXP ans, names;

  /* Initialise the NAG error structure */
  INIT_FAIL(fail);

  /* Perform some initial checks */
  if (LENGTH(ab) != 2) {
    error("ab must be a numeric vector of length 2");
  }
  if (!isLanguage(funct) && !isFunction(funct)) {
    error("funct must be a function or the body of a function");
  }
  if (!isEnvironment(rho)) {
    error("rho must be an environment");
  }
  /* End of parameter checks */

  /* Get initial bounds for the solution */
  a = REAL(ab)[0];
  b = REAL(ab)[1];
  if (a > b) {
    t = a;
    a = b;
    b = t;
  }
  if (!R_FINITE(a) || !R_FINITE(b)) {
    error("ab must be a numeric vector of length 2");
  }

  /* Get required accuracy, or use defaults if not supplied */  
  e1 = REAL(e)[0];
  if (!R_FINITE(e1)) e1 = 0.0;
  if (LENGTH(e) > 1) {
    e2 = REAL(e)[1];
    if (!R_FINITE(e2)) e2 = 0.0;
  } else {
    e2 = 0.0;
  }
  
  /* Get the maximum number of iterations, or use default of 100 if not supplied */
  max_fun = INTEGER(mfun)[0];
  if (max_fun == NA_INTEGER) max_fun = 100;

  /* Allocate memory for the communication array, to hold funct, rho and ifail */
  if (!(p = (void **) malloc(3*sizeof(void *)))) {
    error("Memory allocation error");
  }

  /* Get an error flag for the callback function */
  ifail = 0;

  /* Convert function pointer funct and rho into Nag_Comm structure */
  p[0] = (void *) funct;
  p[1] = (void *) rho;
  p[2] = (void *) &ifail;
  comm.p = p;

  /* Call the NAG routine */
  if (isLanguage(funct)) {
    /* Run version that expects a body of a function */
    e04abc(we04abcL, e1, e2, &a, &b, max_fun, &x, &f, &comm, &fail);
  } else if (isFunction(funct)) {
    /* Run version that expects the function itself */
    e04abc(we04abcF, e1, e2, &a, &b, max_fun, &x, &f, &comm, &fail);
  }

  /* Free up allocated memory */
  free(p);

  /* Check the error returns */
  if (ifail != 0) {
    error("At least one call to the function 'funct' returned a non-numeric value");
  }
  /* Error returns from NAG routine */
  if (fail.code != NE_NOERROR) {
    error(fail.message);
  }

  /* Get the list of results */
  PROTECT(ans = allocVector(VECSXP,3));
  /* Return the estimated point of the minimum (x) */
  SET_VECTOR_ELT(ans,0,nag_d2r(1,&x));
  /* Return the value of F at x (f) */
  SET_VECTOR_ELT(ans,1,nag_d2r(1,&f));
  /* Return the lower and upper bounds for x */
  SET_VECTOR_ELT(ans,2,nag_2d2r(1,&a,1,&b));

  /* Get names for the list of results */
  PROTECT(names = allocVector(STRSXP, 3));
  SET_STRING_ELT(names,0,mkChar("x"));
  SET_STRING_ELT(names,1,mkChar("f"));
  SET_STRING_ELT(names,2,mkChar("ab"));
  setAttrib(ans,R_NamesSymbol,names);

  /* Unprotect the ans and names objects */
  UNPROTECT(2);

  return(ans);
}
  
void NAG_CALL we04abcF(double xc, double *fc, Nag_Comm *comm) {
  /* we04abcF: Wrapper for callback function to E04ABC. For use
     when the user has supplied a function (as opposed to the body of a function)
     Used when calling a NAG C routine
     
     On entry:
     xc   - the point at which the function is to be evaluated
     comm - Communication structure, 
            comm->p[0] - points to the function (or code segment) that
                         is being evaluated. 
            comm->p[1] - points to the environment the function is evaluated in

     On exit:
     fc   - the value of the function at xc
     comm - Communication structure, 
            comm->p[2] - set to 1 if the function returns a non-numeric value
                         In these cases, fc is set to zero.
     
  */
  SEXP funct, rho, R_fcall;
  Integer *ifail;
  void **p;
  p = (void **) comm->p;

  /* Get the pointer to the R function */
  funct = (SEXP) p[0];
  /* Get the pointer to the R environment */
  rho = (SEXP) p[1];

  /* In this case funct points to an actual function that takes a single argument */
  PROTECT(R_fcall = lang2(funct,
                          nag_d2r(1,&xc)));
  
  /* Get the results from evaluating the function in R */
  *fc = REAL(eval(R_fcall, rho))[0];
  
  /* Release R_fcall */
  UNPROTECT(1);

  if (!R_FINITE(*fc)) {
    /* Returned value is not valid. Can't exit NAG routine cleanly, so 
       set the IFAIL flag in COMM, and set returned value to 0
    */
    ifail = (Integer *) p[2];
    *ifail = 1;
    *fc = 0.0;
  }
}

void NAG_CALL we04abcL(double xc, double *fc, Nag_Comm *comm) {
  /* we04abcL: Wrapper for callback function to E04ABC. For use
     when the user has supplied the body of a function (as opposed to the whole function)
     Used when calling a NAG C routine

     On entry:
     xc   - the point at which the function is to be evaluated
     comm - Communication structure, 
            comm->p[0] - points to the function (or code segment) that
                         is being evaluated. 
            comm->p[1] - points to the environment the function is evaluated in

     On exit:
     fc   - the value of the function at xc
     comm - Communication structure, 
            comm->p[2] - set to 1 if the function returns a non-numeric value
                         In these cases, fc is set to zero.
     
  */
  SEXP funct, rho;
  Integer *ifail;
  void **p;
  p = (void **) comm->p;

  /* Get the pointer to the R function */
  funct = (SEXP) p[0];
  /* Get the pointer to the R environment */
  rho = (SEXP) p[1];

  /* Makes xc available to the function as a variable named x */
  defineVar(install("x"), nag_d2r(1,&xc), rho);

  /* Get the results from evaluating the function in R */
  *fc = REAL(eval(funct,rho))[0];

  if (!R_FINITE(*fc)) {
    /* Returned value is not valid. Can't exit NAG routine cleanly, so 
       set the IFAIL flag in COMM, and set returned value to 0
    */
    ifail = (Integer *) p[2];
    *ifail = 1;
    *fc = 0.0;
  }
}
